/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "blobs.h"

void blobFollowOrders();
void blobDoOrderFollow();

void blobJumping()
{
	if (self->flags & EF_ONGROUND)
	{
		self->action = blobDoOrderFollow;
	}
	
	self->thinkTime = 1;
}

bool blobJumpUp()
{
	if (!(player->flags & EF_ONGROUND))
	{
		return false;
	}
	
	Unit *unit = (Unit*)self;
	
	float zDist = (player->position.z - unit->position.z);
	
	if ((zDist > 3) && (zDist < 50))
	{
		unit->velocity = player->position;
		unit->velocity -= unit->position;
		unit->velocity.normalize();
		
		unit->velocity.z = 1.25;
				
		unit->action = &blobJumping;
				
		Math::removeBit(&unit->flags, EF_ONGROUND);
		
		return true;
	}
	
	return false;
}

bool blobJumpGap()
{
	Unit *unit = (Unit*)self;
	
	Vector start, end, velocity;
	bool canMakeJump = false;
	
	if (unitIsAtEdge(unit))
	{
		if (fabs(player->position.z - self->position.z) > 12)
		{
			return false;
		}
		
		// attempt to jump to next ledge
		
		if (unit->flags & EF_ONGROUND)
		{
			velocity = player->position;
			velocity -= unit->position;
			velocity.normalize();
		}
		
		canMakeJump = false;
		
		start = unit->position;
		
		for (int i = 0 ; i < 150 ; i += 6)
		{
			start += (velocity * 6);
			end = start;
			end.z -= 6;
			
			if (!bsp->canSightTarget(start, end))
			{
				canMakeJump = true;
				break;
			}
			
			if (bbManager->canSightSolid(unit, start, end))
			{
				canMakeJump = true;
				break;
			}
		}
		
		if (canMakeJump)
		{
			unit->velocity = player->position;
			unit->velocity -= unit->position;
			unit->velocity.normalize();
			unit->velocity = unit->velocity * 1.25;
			unit->velocity.z = 1.25;
				
			unit->action = &blobJumping;
				
			Math::removeBit(&unit->flags, EF_ONGROUND);
		}
	}
	
	return canMakeJump;
}

bool blobAttemptJump()
{
	return (blobJumpUp() || blobJumpGap());
}

void blobCircleTarget()
{
	self->velocity.x = Math::rrand(-3, 3);
	self->velocity.y = Math::rrand(-3, 3);
	
	self->velocity.x /= 10;
	self->velocity.y /= 10;
}

void blobDodgeTarget()
{
	self->velocity.x = Math::rrand(-2, 2);
	self->velocity.y = Math::rrand(-2, 2);
	
	self->velocity.x /= Math::rrand(7, 10);
	self->velocity.y /= Math::rrand(7, 10);
}

void blobsRunToBob()
{
	float z;
	
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if (unit == player)
		{
			continue;
		}
		
		z = unit->velocity.z;
		
		faceLocation(unit, player->position);
	
		unit->velocity = player->position;
		unit->velocity -= unit->position;
		unit->velocity.normalize();
	
		float speed = 1.0;
	
		unit->velocity.x /= speed;
		unit->velocity.y /= speed;
		unit->velocity.z = z;
	
		unit->thinkTime = 9999;
	}
}

void blobStayNearBob()
{
	float distance = Math::getDistance(self->position.x, self->position.y, player->position.x, player->position.y);
	
	// too far away!! Fall back!!
	if (distance > 150)
	{
		float z = self->velocity.z;
		
		self->velocity = player->position;
		self->velocity -= self->position;
		self->velocity.normalize();
	
		self->velocity.x /= 3.0;
		self->velocity.y /= 3.0;
		self->velocity.z = z;
	}
}

void blobWander()
{
	self->velocity.x = Math::rrand(-20, 20);
	self->velocity.y = Math::rrand(-20, 20);
	self->velocity.x /= 100;
	self->velocity.y /= 100;

	float angle = atan2(self->velocity.y, self->velocity.x);
	self->rotation.x = 90 + toDegrees(angle);
		
	self->thinkTime = Math::rrand(200, 400);
}

void miaBlobThink()
{
	if (!bsp->canSightTarget(self->position, player->position))
	{
		if ((rand() % 3) == 0)
		{
			blobWander();
		}
		else
		{
			self->velocity.x = 0;
			self->velocity.y = 0;
			
			self->thinkTime = Math::rrand(200, 400);
		}
		
		return;
	}
	
	self->velocity.x = 0;
	self->velocity.y = 0;
	
	faceLocation((Unit*)self, player->position);
	
	float distance = Math::getDistance(self->position, player->position);
	
	if (distance <= 10)
	{
		if (game->hasItem("TELETAG"))
		{
			game->setGamePlayMessage(localeManager->translate("HINT_USE_TELETAG"));
		}
		else
		{
			game->setGamePlayMessage(localeManager->translate("HINT_NEED_TELETAG"));
		}
	}
}

void blobFollowBob()
{
	Vector sightPosition = self->position;
	
	sightPosition.z += Math::rrand(0, 25);
	
	float distance = Math::getDistance(self->position.x, self->position.y, player->position.x, player->position.y);
	
	if (!bsp->canSightTarget(sightPosition, player->position))
	{
		// Bob is very close... Try and get to him.
		if (distance < 500)
		{
			self->thinkTime = Math::rrand(1, 5);
			self->action = blobDoOrderFollow;
			return;
		}
		
		if (rand() % 5 > 0)
		{
			self->velocity.x = self->velocity.y = 0;
			self->thinkTime = Math::rrand(200, 400);
			self->action = blobDoOrderFollow;
			return;
		}
		
		blobWander();
		
		self->action = blobDoOrderFollow;
		
		return;
	}
	
	self->action = &blobFollowOrders;
	
	faceLocation((Unit*)self, player->position);
	
	if (game->currentOrders != ORDERS_RETREAT)
	{
		if (distance < 30)
		{
			if (!blobJumpUp())
			{
				self->velocity.x = 0;
				self->velocity.y = 0;
			}
			return;
		}
	}
	
	float z = self->velocity.z;
	
	self->velocity = player->position;
	self->velocity -= self->position;
	self->velocity.normalize();
	
	float speed = 8.0 - (distance / 10);
	
	if (game->currentOrders != ORDERS_RETREAT)
	{
		Math::limit(&speed, 1.2, 3.0);
	}
	else
	{
		Math::limit(&speed, 0.6, 1.5);
	}
	
	self->thinkTime = Math::rrand(5, 10);
	
	self->velocity.x /= speed;
	self->velocity.y /= speed;
	self->velocity.z = z;
	
	blobAttemptJump();
}

bool blobSightTarget(Unit *unit, int range)
{
	unit->target = NULL;
	
	float distance = 0;
	float targetDistance = range;
	
	Vector sightPosition = unit->position;
	
	sightPosition.z += Math::rrand(0, 25);
	
	for (Unit *enemy = (Unit*)entityManager->enemyList.getFirstElement() ; enemy != NULL ; enemy = (Unit*)enemy->next)
	{
		if ((!enemy->canBeDamaged()) ||( enemy->health < 1) || (enemy->flags & EF_STATIC))
		{
			continue;
		}
		
		distance = Math::getDistance(unit->position, enemy->position);
		
		if (distance < targetDistance)
		{
			if (!unitCanSeeTarget(unit, enemy, sightPosition))
			{
				continue;
			}
			
			targetDistance = distance;
			unit->target = enemy;
		}
	}
	
	return (unit->target != NULL);
}

void goodBlobAttack()
{
	Unit *unit = (Unit*)self;
	
	if (unit->target->flags & EF_DEAD)
	{
		unit->custom = 0;
		unit->action = &blobFollowOrders;
		return;
	}
	
	faceTarget(unit);
	
	if (unit->reload > 0)
	{
		return;
	}
	
	unit->thinkTime = 0;
	unit->action = &goodBlobAttack;
	
	if (unit->currentWeapon->type == WEP_UZI)
	{
		attackTargetScatter(unit);
	}
	else
	{
		attackTarget(unit);
	}
	
	if ((--unit->custom == 0) || (unit->target->health <= 0))
	{
		unit->thinkTime = Math::rrand(50, 100);
		unit->action = &blobFollowOrders;
	}
}

void goodBlobSetFireRate()
{
	Unit *unit = (Unit*)self;
	
	switch (unit->currentWeapon->type)
	{
		case WEP_PISTOL:
			unit->custom = Math::rrand(1, 2);
			break;
			
		case WEP_UZI:
			unit->custom = Math::rrand(4, 8);
			break;
	}
}

void goodBlobPreAttack()
{
	blobStayNearBob();
	
	switch (rand() % 6)
	{
		case 0:
			// make blob move around target
			blobDodgeTarget();
			goodBlobSetFireRate();
			self->action = goodBlobAttack;
			break;
			
		case 1:
			// add some circle strafing code
			blobCircleTarget();
			goodBlobSetFireRate();
			self->action = goodBlobAttack;
			break;
			
		case 2:
			// make blob stay stationary only once!
			self->velocity.x = self->velocity.y = 0;
			goodBlobSetFireRate();
			self->action = goodBlobAttack;
			break;
			
		case 3:
		case 4:
		case 5:
			blobDodgeTarget();
			self->thinkTime = Math::rrand(50, 100);
			break;
	}
}

bool blobSightHealth()
{
	if (self->health == self->maxHealth)
	{
		return false;
	}
	
	Item *item;
	ItemDef *itemDef;
	
	float distance;
	
	for (Entity *ent = (Entity*)entityManager->itemList.getFirstElement() ; ent != NULL ; ent = (Entity*)ent->next)
	{
		if (fabs(self->position.z - ent->position.z) > 6)
		{
			continue;
		}
		
		item = (Item*)ent;
		itemDef = (ItemDef*)item->definition;
		
		if (itemDef->itemType == IT_MEDICAL)
		{
			distance = Math::getDistance(self->position.x, self->position.y, item->position.x, item->position.y);
			
			if (distance > 75)
			{
				continue;
			}
			
			if (unitCanSeeTarget((Unit*)self, ent, self->position))
			{
				float oldZ = self->velocity.z;
				
				self->velocity = ent->position;
				self->velocity -= self->position;
				self->velocity.normalize();
				
				float speed = 8.0 - (distance / 10);
				Math::limit(&speed, 1.5, 3.0);
		
				self->velocity.x /= speed;
				self->velocity.y /= speed;
				self->velocity.z = oldZ;
				
				faceLocation((Unit*)self, item->position);
				
				debug(("%s is going for health...\n", self->getName()));
			
				return true;
			}
		}
	}
	
	return false;
}

void goodBlobGetItem()
{
	if (--self->custom == 0)
	{
		self->action = &blobFollowOrders;
	}
}

void blobDoOrderFollow()
{
	if (blobSightHealth())
	{
		self->action = &goodBlobGetItem;
		self->custom = 100;
	}
	else if (blobSightTarget((Unit*)self, 226))
	{
		faceTarget((Unit*)self);
		self->action = &goodBlobPreAttack;
	}
	else
	{
		self->action = &blobFollowBob;
	}
}

void blobDoOrderWait()
{
	self->velocity.x = 0;
	self->velocity.y = 0;
	
	if (blobSightHealth())
	{
		self->action = &goodBlobGetItem;
		self->custom = 100;
	}
	else if (blobSightTarget((Unit*)self, 226))
	{
		faceTarget((Unit*)self);
		self->action = &goodBlobPreAttack;
	}
	else
	{
		faceLocation((Unit*)self, player->position);
	}
	
	self->thinkTime = Math::rrand(1, 2);
}

void blobDoOrderRetreat()
{
	self->action = &blobFollowBob;
}

void blobWait()
{
	self->thinkTime = 450;
	
	self->action = &blobFollowOrders;
}

void resetBlobOrders()
{
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		unit->completedOrders = false;
	}
}

void blobDoOrderUse()
{
	self->thinkTime = Math::rrand(1, 3);
	
	self->velocity.x = 0;
	self->velocity.y = 0;
	
	Entity *closest = NULL;
	float distance;
	float lastDistance = 9999;
	
	for (Entity *entity = (Entity*)entityManager->elementList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if (entity->entityType != ET_SWITCH)
		{
			continue;
		}
		
		distance = Math::getDistance(entity->position.x, entity->position.y, self->position.x, self->position.y);
		
		if (distance > 140)
		{
			continue;
		}
		
		if (distance < lastDistance)
		{
			closest = entity;
			lastDistance = distance;
		}
	}
	
	if (closest == NULL)
	{
		return;
	}
	
	// walk the Blob to the feature
	
	float z = self->velocity.z;
		
	self->velocity = closest->position;
	self->velocity -= self->position;
	self->velocity.normalize();
	
	self->velocity = (self->velocity * 0.85);
	
	self->velocity.z = z;
	
	faceLocation((Unit*)self, closest->position);
	
	if (Math::getDistance(closest->position.x, closest->position.y, self->position.x, self->position.y) <= 8)
	{
		self->velocity.x = self->velocity.y = 0;
		
		switch (closest->entityType)
		{
			case ET_SWITCH:
				handleSwitchUse((Switch*)closest);
				((Unit*)self)->completedOrders = true;
				self->thinkTime = 50;
				break;
		}
		
		return;
	}
}

void blobFollowOrders()
{
	switch (game->currentOrders)
	{
		case ORDERS_FOLLOW:
			blobDoOrderFollow();
			break;
			
		case ORDERS_RETREAT:
			blobDoOrderRetreat();
			break;
			
		case ORDERS_WAIT:
			blobDoOrderWait();
			break;
			
		case ORDERS_USE:
			if (!((Unit*)self)->completedOrders)
			{
				blobDoOrderUse();
			}
			else
			{
				blobDoOrderWait();
			}
			break;
	}
}

Unit *findTeeka()
{
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if (unit->definition->type == BLOB_TEEKA)
		{
			return unit;
		}
	}
	
	printf("WARNING: findTeeka() returning NULL!\n");
	
	return NULL;
}

void teekaVanish()
{
	debug(("teekaVanish()\n"));
	
	game->teekaAvailableTimer = 1000;
	
	addTeleportParticles(self->position);
	
	Math::addBit(&self->flags, EF_VANISHED);

	audio->playSound(SND_TELEPORT3, CH_SPAWN);
	
	self->action = NULL;
}

void teekaReappear()
{
	int x = Math::rrand(-2, 2);
	int y = Math::rrand(-2, 2);
	
	if (x == 0) x = -1;
	if (y == 0) y = -1;
	
	x *= 12;
	y *= 12;
	
	self->position = player->position;
	self->position.x += x;
	self->position.y += y;
	self->position.z += 18;
	
	if (!unitCanSpawnInLocation(NULL, self->position))
	{
		game->teekaAvailableTimer = 1;
		return;
	}
	
	debug(("teekaReappear()\n"));
	
	addTeleportParticles(self->position);
	
	Math::removeBit(&self->flags, EF_VANISHED);

	audio->playSound(SND_TELEPORT3, CH_SPAWN);
	
	self->action = blobFollowOrders;
	
	((Unit*)self)->liquidLevel = 0;
	
	self->contentType = CONTENTS_SOLID;
	
	self->velocity.set(0, 0, 0);
}

void addBlob(Properties *props)
{
	int type = entityManager->getEntityType(props);
	
	if ((game->changingLevel) && (type == BLOB_BOB))
	{
		player->position = props->getVector("position");
		return;
	}
	
	Unit *unit = entityManager->spawnBlob(type);
	
	unit->die = &generalBlobDie;
	unit->walk = &blobFollowOrders;
	unit->action = unit->walk;
	
	switch (unit->definition->type)
	{
		case BLOB_BOB:
			unit->name = "Bob";
			player = unit;
			player->action = NULL;
		
			debug(("Player assigned to Blob '%s', id %d\n", unit->getName(), unit->getUniqueId()));
			break;
			
		case BLOB_TEEKA:
			unit->name = "Teeka";
			unit->currentWeapon = &entityManager->weaponDef[game->teekaWeapon];
			break;
			
		case BLOB_MIA:
			unit->setName(randomNameGenerator->getRandomName());
			unit->action = &miaBlobThink;
			break;
			
		default:
			unit->setName(randomNameGenerator->getRandomName());
			break;
	}
	
	unit->load(props);
	
	debug(("addBlob() [%s %d %.2f %.2f]\n", unit->getName(), unit->definition->type, unit->health, unit->shield));
}

void doBlobs()
{
	bool helpless;
	float fallSpeed;
	
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		self = unit;
		
		bbManager->removeBox(unit);
		
		// Handle Teeka first...
		if (unit->definition->type == BLOB_TEEKA)
		{
			if (unit->flags & EF_VANISHED)
			{
				if (player->health > 0)
				{
					game->teekaAvailableTimer -= engine->getTimeDifference(TD_LOGIC);
				}
				
				Math::limit(&game->teekaAvailableTimer, 0, 1000);
				
				if (game->teekaAvailableTimer == 0)
				{
					teekaReappear();
				}
				else
				{
					continue;
				}
			}
		}
		
		if (unit->health <= 0)
		{
			if (unit->flags & EF_DEAD)
			{
				if ((unit == player) || (unit->definition->type == BLOB_TEEKA))
				{
					game->setMissionStatus(MISSION_FAILED);
					unit->referenced = false;
					continue;
				}
			}
			else if (!(unit->flags & EF_DYING))
			{
				unit->die();
				
				if (!unit->dead)
				{
					game->setInfoMessage(INFO_BAD, localeManager->translate("INFO_KILLED"), unit->getName());
					fireTriggers(unit->getName(), TRIGGER_TYPE_ENTITY_KILLED);
					unit->dead = true;
				}
				
				Math::addBit(&unit->flags, EF_DYING);
				Math::addBit(&unit->flags, EF_BOUNCES);
			}
		}
		
		// we can do this here because we've already checked it
		unit->referenced = false;
		
		helpless = (unit->helpless > 0);

		if ((!helpless) || (unit->health < 1))
		{
			if (unit->performNextAction(engine->getTimeDifference(TD_LOGIC)))
			{
				if (unit->action != NULL)
				{
					unit->action();
				}
			}
		}

		unit->think(engine->getTimeDifference(TD_LOGIC));
		
		if ((unit != player) && (unit->definition->type != BLOB_TEEKA))
		{
			if (!(unit->contentType & CONTENTS_SOLID))
			{
				if ((unit->contentType & CONTENTS_WATER) && (unit->liquidLevel > 1))
				{
					if (unit->health > 0)
					{
						unit->applyDamage(unit->health);
					}
				}
				else if (unit->contentType & (CONTENTS_SLIME|CONTENTS_LAVA))
				{
					if (unit->health > 0)
					{
						unit->applyDamage(unit->health);
					}
				}
			}
		}
		else if ((unit->definition->type == BLOB_TEEKA) && (!(unit->flags & EF_VANISHED)))
		{
			if ((unit->contentType & (CONTENTS_SLIME|CONTENTS_LAVA)) || (unit->liquidLevel == 2))
			{
				teekaVanish();
			}
		}

		// was helpless but now isn't - Stop sliding!
		if ((helpless) && (unit->helpless <= 0))
		{
			if ((unit->flags & EF_ONGROUND) || (!(unit->contentType & CONTENTS_SOLID)))
			{
				unit->rotation.set(0, 0, 0);
				Math::removeBit(&unit->flags, EF_BOUNCES);
			}
			else
			{
				unit->helpless = 3;
			}
		}

		unit->applyGravity(engine->getTimeDifference(TD_LOGIC));
		
		if ((unit != player) && (unit->flags & EF_ONGROUND))
		{
			keepUnitAwayFromEdges(unit);
		}
		
		fallSpeed = unit->realVelocity.z;
		
		if (fallSpeed <= -5)
		{
			unit->helpless = 5;
			Math::addBit(&unit->flags, EF_BOUNCES);
		}
		
		moveEntity(unit);
		
		/*
		unit->prepareToMove(engine->getTimeDifference(TD_LOGIC));
		
		if (bsp->checkToWorld(unit))
		{
			bbManager->checkToOthers(unit);
		}
		
		//unit->move(engine->getTimeDifference(TD_LOGIC));
		unit->position = tracer->currentPosition;
		unit->velocity = unit->realVelocity;
		
		if ((unit->flags & EF_ONGROUND) && (!(unit->flags & EF_BOUNCES)))
		{
			unit->velocity.z = 0;
		}
		*/
		
		//if (tracer->outputContent != unit->contentType)
		{
			unitContentChanged(unit);
		}
		
		// take damage from falling far but not if you hit liquid
		if ((fallSpeed <= -5) && (unit->velocity.z > -5) && (unit->contentType & CONTENTS_SOLID))
		{
			fallSpeed = fabs(fallSpeed * 3);
			unit->applyDamage(fallSpeed);
			debug(("%s took %.2f damage from fall\n", unit->getName(), fallSpeed));
			
			if (unit == player)
			{
				game->alterStat("TOTAL_DAMAGE_TAKEN", fallSpeed);
			}
		}
		
		if ((unit->position.z < bsp->minCords.z) && (unit->health > 0))
		{
			printf("WARNING: Blob '%s' fell out of map at %s\n", unit->getName(), unit->position.toString());
			unit->health = -100;
		}
		
		bbManager->addBox(unit);
	}
}
