/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "bullets.h"

void assignBulletActionType(Bullet *bullet);

void doBulletUnitHit(Unit *unit, int damage)
{
	Bullet *bullet = (Bullet*)self;
	
	bool wasAlive = (unit->health > 0);
	
	float shield = unit->shield;
	
	unit->applyDamage(damage);
	
	if (unit == player)
	{
		game->alterStat("TOTAL_DAMAGE_TAKEN", damage);
	}
	
	if ((unit->shield == -2.5) && (shield > 0))
	{
		addShieldDestroyedParticles(unit);
		unit->shieldHit = 0;
	}
	
	if (bullet->owner != NULL)
	{
		if ((wasAlive) && (unit->health <= 0) && (unit->entityType == ET_BIOMECH))
		{
			switch (bullet->owner->entityType)
			{
				case ET_BLOB:
					if (bullet->owner == player)
					{
						game->alterStat("ENEMIES_KILLED", 1);
					}
					else
					{
						game->alterStat("ENEMIES_KILLED_TEEKA", 1);
					}
					break;
				
				case ET_BIOMECH:
					game->alterStat("ENEMIES_KILLED_FRIENDLY", 1);
					break;
			}
		}
	}
				
	game->alterStat("BULLETS_HIT", 1);
		
	if (unit == player)
	{
		game->alterStat("TOTAL_DAMAGE_TAKEN", damage);
	}
	
	// first check for combo hits
	if ((unit->health <= 0) && (unit->flags & EF_DYING))
	{
		if (!(unit->flags & EF_STATIC))
		{
			throwUnit(unit, bullet->velocity);
			Math::removeBit(&unit->flags, EF_DYING);
		}
		
		if ((bullet->owner == player) || (bullet->owner->definition->type == BLOB_TEEKA))
		{
			game->comboTimer = 150;
			game->comboHits++;
		}
		
		return;
	}
	
	if ((unit->maxShield == 0) || (unit->shield <= 0))
	{
		if (!(unit->flags & EF_EXPLODES))
		{
			addBloodParticles(bullet);
			audio->playSound(SND_HIT, CH_ANY);
		}
		else
		{
			addSparkParticle(bullet);
			audio->playSound(SND_CLANG, CH_ANY);
		}
	}
	else
	{
		if (unit->shield > 0)
		{
			addShieldHitParticles(bullet->position);
			audio->playSound(SND_ELECTRIC_SPARK1, CH_ANY);
		}
	}
	
	if (unit->health <= 0)
	{
		// force of impact - should be overridden in die() if needed
		if (!(unit->flags & EF_STATIC))
		{
			unit->velocity = bullet->velocity;
			unit->velocity.x /= 4;
			unit->velocity.y /= 4;
			unit->velocity.z = 0.25;
		}
		
		if (bullet->owner == player)
		{
			game->comboTimer = 150;
			game->comboHits++;
		}
	}
	else
	{
		unit->velocity.x += (bullet->velocity.x * 0.05);
		unit->velocity.y += (bullet->velocity.y * 0.05);
		unit->reload += 10;
	}
	
	if ((unit != player) && (unit->target == NULL) && (!(unit->flags & EF_STATIC)))
	{
		unit->target = (Unit*)bullet->owner;
		unit->thinkTime = 0;
	}
}

void doBulletDamage(Entity *other, int damage)
{
	if ((self->health == 0) || (other == self->owner))
	{
		return;
	}
	
	Bullet *bullet = (Bullet*)self;
	
	bullet->health = 0;
	
	if (other != NULL)
	{
		if (other->entityType == ET_BULLET)
		{
			return;
		}
		
		if ((other->entityType != ET_BIOMECH) && (bullet->owner == player))
		{
			game->alterStat("BULLETS_MISSED", 1);
		}
		
		Unit *unit;
		
		switch (other->entityType)
		{
			case ET_BLOB:
			case ET_BIOMECH:
				unit = (Unit*)other;
				doBulletUnitHit(unit, damage);
				if (unit->totalCurrentDamage >= 5)
				{
					throwUnit(unit, bullet->velocity);
				}
				break;
				
			case ET_WINDOW:
				other->health -= damage;
				addSparkParticle(bullet);
				break;
				
			case ET_BARRIER:
				addSparkParticle(bullet);
				audio->playSound(SND_ELECTRIC_SPARK1, CH_ANY);
				break;
				
			case ET_STRUCTURE:
				if (!(other->flags & EF_IMMORTAL))
				{
					other->health -= damage;
					addSparkParticle(bullet);
				}
				else
				{
					addBulletHitParticles(bullet);
					audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY);
				}
				break;
				
			default:
				addBulletHitParticles(bullet);
				audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY);
				break;
		}
	}
	else
	{
		addBulletHitParticles(bullet);
		audio->playSound(Math::rrand(SND_RICO1, SND_RICO2), CH_ANY);
		
		if (bullet->owner == player)
		{
			game->alterStat("BULLETS_MISSED", 1);
		}
	}
}

void bulletHitSingleDamage(Entity *other)
{
	doBulletDamage(other, 1);
}

void bulletHitDamage(Entity *other)
{
	Bullet *bullet = (Bullet*)self;
	
	doBulletDamage(other, ((WeaponDef*)bullet->definition)->damage);
}

void grenadeHit(Entity *other)
{
	if (other == NULL)
	{
		if (self->realVelocity.z > EPSILON)
		{
			audio->playSound(SND_METAL_BOUNCE, CH_ANY);
		}
		else
		{
			Math::removeBit(&self->flags, EF_BOUNCES);
		}
		
		return;
	}
	
	if (self->owner == other)
	{
		return;
	}
	
	switch (other->entityType)
	{
		case ET_DOOR:
		case ET_SWITCH:
			audio->playSound(SND_METAL_BOUNCE, CH_ANY);
			break;
			
		case ET_STRUCTURE:
			if (other->flags & EF_IMMORTAL)
			{
				audio->playSound(SND_METAL_BOUNCE, CH_ANY);
			}
			else
			{
				self->health = 0;
			}
			break;
			
		default:
			self->health = 0;
			break;
	}
}

void bulletDieExplode()
{
	Bullet *bullet = (Bullet*)self;
	
	doExplosion(bullet->position, ((WeaponDef*)bullet->definition)->damage, bullet->owner);
}

void addNormalBullet(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	destination -= origin;
	destination.normalize();
	destination.x *= 2.5;
	destination.y *= 2.5;
	destination.z *= 2.5;
	
	Bullet *bullet = entityManager->spawnBullet();
	bullet->setOwner(owner);
	bullet->definition = definition;
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 200;
	bullet->rotation = owner->rotation;
	
	if (definition->flags & WF_BOUNCES)
	{
		bullet->flags += EF_BOUNCES;
	}
	
	assignBulletActionType(bullet);
	
	if (owner->entityType == ET_BLOB)
	{
		bullet->color = GLColor::white;
	}
	else
	{
		bullet->color = GLColor::purple;
	}
}

void addPellet(Unit *owner, WeaponDef *definition, Vector origin, Vector destination)
{
	// alter the destination so we cause the bullets to spray out
	int adjX = 25;
	int adjY = 25;
	
	if (fabs(origin.x - destination.x) > 25)
	{
		adjX = (int)fabs(origin.x - destination.x);
	}
	
	if (fabs(origin.x - destination.y) > 25)
	{
		adjY = (int)fabs(origin.y - destination.y);
	}
	
	float x = Math::rrand(-adjX, adjX);
	float y = Math::rrand(-adjY, adjY);
	float z = Math::rrand(-5, 5);
	
	x /= 3;
	y /= 3;
	z /= 3;
	
	destination.x += x;
	destination.y += y;
	destination.z += z;
	
	destination -= origin;
	destination.normalize();
	destination.x *= 2.5;
	destination.y *= 2.5;
	destination.z *= 2.5;
	
	Bullet *bullet = entityManager->spawnBullet();
	bullet->setOwner(owner);
	bullet->definition = definition;
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->health = 200;
	bullet->rotation = owner->rotation;
	bullet->flags = 0;
	
	assignBulletActionType(bullet);
	
	if (owner->entityType == ET_BLOB)
	{
		bullet->color = GLColor::white;
	}
	else
	{
		bullet->color = GLColor::purple;
	}
}

void fireWeaponBeam(Unit *owner, WeaponDef *weaponDef, Vector origin, Vector destination)
{
}

void throwGrenade(Unit *owner, WeaponDef *weaponDef, Vector origin, Vector destination)
{
	Bullet *bullet = entityManager->spawnBullet();
	bullet->setOwner(owner);
	bullet->definition = &entityManager->weaponDef[WEP_GRENADE];
	bullet->position = origin;
	bullet->velocity = destination;
	bullet->rotation = owner->rotation;
	bullet->flags = EF_BOUNCES;
	
	bullet->velocity = destination;
	bullet->velocity -= origin;
	
	float steps;
	steps = (fabs(origin.x - destination.x));
	steps = max(steps, (fabs(origin.y - destination.y)));
	
	float adjust = 0.5;
	
	if (owner != player)
	{
		float extra = rand() % 25;
		extra /= 100;
		adjust += extra;
	}
	
	bullet->velocity.x /= steps;
	bullet->velocity.y /= steps;
	bullet->velocity.z = (steps / 2) * (GRAVITY * adjust);
	
	bullet->health = 100 + steps;
	
	assignBulletActionType(bullet);
	
	owner->reload = 100;
}

void throwClusterBombs(Unit *owner, WeaponDef *weaponDef, Vector origin, Vector destination)
{
	Bullet *bullet;
	float steps;
	float diff;
	
	for (int i = 0 ; i < 5 ; i++)
	{
		bullet = entityManager->spawnBullet();
		bullet->setOwner(owner);
		bullet->definition = &entityManager->weaponDef[WEP_GRENADE_LAUNCHER];
		bullet->position = origin;
		bullet->velocity = destination;
		bullet->health = 9999;
		bullet->rotation = owner->rotation;
		bullet->flags = entityManager->weaponDef[WEP_GRENADE_LAUNCHER].flags;
		
		bullet->velocity = destination;
		bullet->velocity.x += Math::rrand(-10, 10);
		bullet->velocity.y += Math::rrand(-10, 10);
		bullet->velocity -= origin;
		
		steps = (fabs(origin.x - destination.x));
		steps = max(steps, (fabs(origin.y - destination.y)));
		
		diff = Math::rrand(90, 110);
		diff /= 100;
		
		steps *= diff;
		
		bullet->velocity.x /= steps;
		bullet->velocity.y /= steps;
		bullet->velocity.z = (steps / 2) * 0.01;
		
		assignBulletActionType(bullet);
	}
	
	owner->reload = 200;
}

void assignBulletActionType(Bullet *bullet)
{
	switch (bullet->definition->type)
	{
		case WEP_SHOTGUN:
			bullet->touch = bulletHitSingleDamage;
			break;
		
		case WEP_GRENADE:
		case WEP_BIOMECH_GRENADE:
		case WEP_SMOKE_GRENADE:
		case WEP_STUN_GRENADE:
		case WEP_CHAFF_GRENADE:
		case WEP_NAPALM_GRENADE:
			bullet->touch = grenadeHit;
			bullet->die = bulletDieExplode;
			break;
		
		default:
			bullet->touch = bulletHitDamage;
			break;
	}
}

void addBullet(Properties *props)
{
	Bullet *bullet = entityManager->spawnBullet();
	
	int type = props->getInt("definitionType", 0);
	
	bullet->definition = &entityManager->weaponDef[type];
	
	bullet->load(props);
	
	assignBulletActionType(bullet);
}

void drawBullet(Bullet *bullet)
{
	//drawBoundingBox(bullet);
	
	if (!(bullet->definition->flags & WF_EXPLODES))
	{
		glEnable(GL_TEXTURE_2D);
		
		glPushMatrix();
		{
			glTranslatef(bullet->position.x, bullet->position.y, bullet->position.z);
			
			if (((Unit*)bullet->owner)->entityType == ET_BLOB)
			{
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("Bullet")->data);
				glColor4fv(bullet->color.color);
			}
			else
			{
				glDisable(GL_TEXTURE_2D);
				glColor4fv(GLColor::getRandomColor().color);
			}
			
			glRotatef(bullet->rotation.x, 0.0, 0.0, 1.0);
			glRotatef(bullet->rotation.y, 0.0, 1.0, 0.0);
			glRotatef(bullet->rotation.z, 1.0, 0.0, 0.0);
			
			glCallList(modelManager->getSimpleModel("Bullet")->data);
		}
		glPopMatrix();
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
		
		glPushMatrix();
		{
			glColor4fv(GLColor::green.color);
			
			glTranslatef(bullet->position.x, bullet->position.y, bullet->position.z);
			
			gluSphere(graphics->quadric, 1, 4, 8);
		}
		glPopMatrix();
	}
}

/*
 * Will show blood if required
*/
void doBulletHitEffect(Bullet *bullet, Unit *unit)
{
	if (!(unit->flags & EF_EXPLODES))
	{
		addBloodParticles(bullet);
	}
	else
	{
		addSparkParticle(bullet);
	}
	
	game->alterStat("BULLETS_HIT", 1);
}

void doBullets()
{
	glDisable(GL_BLEND);
	
	for (Bullet *bullet = (Bullet*)entityManager->bulletList.getFirstElement() ; bullet != NULL ; bullet = (Bullet*)bullet->next)
	{
		self = bullet;
		
		bbManager->removeBox(bullet);
		
		bullet->owner->referenced = true;
		
		if (withinViewableRange(bullet->position, 1))
		{
			drawBullet(bullet);
		}
		
		bullet->applyGravity(engine->getTimeDifference(TD_LOGIC));
		
		/*
		bullet->prepareToMove(engine->getTimeDifference(TD_LOGIC));
		
		if (bsp->checkToWorld(bullet))
		{
			bbManager->checkToOthers(bullet);
		}
		
		bullet->move(engine->getTimeDifference(TD_LOGIC));
		*/
		moveEntity(bullet);
		
		bullet->health -= engine->getTimeDifference(TD_LOGIC);
		
		if ((tracer->outputContent & CONTENTS_WATER) && (!(bullet->contentType & CONTENTS_WATER)))
		{
			audio->playSound(SND_BULLET_HIT_WATER, CH_EFFECT1);
		}
		
		bullet->contentType = tracer->outputContent;
		
		if (bullet->contentType & CONTENTS_WATER)
		{
			addBubbleParticle(bullet->position);
		}
		
		if (((WeaponDef*)bullet->definition)->hasTrail)
		{
			addTrailParticle(bullet->position, 1, GLColor::green);
		}
		
		if (bullet->health <= 0)
		{
			if (bullet->die != NULL)
			{
				bullet->die();
			}
			
			bullet = (Bullet*)bullet->previous;
			entityManager->bulletList.remove(bullet->next);
			continue;
		}
		
		bbManager->addBox(bullet);
	}
}
