/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "decoration.h"

void addGlassShard(Triangle *triangle)
{
	Decoration *decoration = entityManager->spawnDecoration(DT_SHARD);
	
	decoration->health = Math::rrand(100, 300);
	decoration->velocity.x = Math::rrand(-3, 3);
	decoration->velocity.y = Math::rrand(-3, 3);
	decoration->velocity.z = Math::rrand(0, 3);
	decoration->size = 1;
	
	decoration->custom = Math::rrand(100, 300);
	decoration->custom /= 1000;
	
	decoration->velocity.x /= 10;
	decoration->velocity.y /= 10;
	decoration->velocity.z /= 10;
	
	Math::addBit(&decoration->flags, EF_DYING);
	
	decoration->triangle = *triangle;
}

void throwFleshChunk(Vector position, bool underWater)
{
	static float x, y;
	
	Decoration *decoration = entityManager->spawnDecoration(DT_FLESH);
	
	decoration->position = position;
	decoration->health = 300;
	decoration->velocity.x = Math::rrand(-25, 25);
	decoration->velocity.y = Math::rrand(-25, 25);
	
	if (!underWater)
	{
		decoration->velocity.z = Math::rrand(2, 10);
		
		decoration->size = Math::rrand(25, 100);
		decoration->size /= 100;
	}
	else
	{
		x = Math::rrand(-300, 300);
		y = Math::rrand(-300, 300);
	
		x /= 100;
		y /= 100;
		
		decoration->position.x += x;
		decoration->position.y += y;
		
		decoration->size = 0.25;
		
		decoration->velocity.x = Math::rrand(-5, 5);
		decoration->velocity.y = Math::rrand(-5, 5);
	}
	
	decoration->velocity.x /= 100;
	decoration->velocity.y /= 100;
	decoration->velocity.z /= 10;
	
	Math::addBit(&decoration->flags, EF_DYING);
	
	if (underWater)
	{
		Math::addBit(&decoration->flags, EF_WEIGHTLESS);
	}
}

void throwExplosionDebris(Vector position)
{
	Decoration *decoration = entityManager->spawnDecoration(DT_FIRE);
	
	decoration->position  = position;
	decoration->health = Math::rrand(25, 150);
	decoration->velocity.x = Math::rrand(-100, 100);
	decoration->velocity.y = Math::rrand(-100, 100);
	decoration->velocity.z = Math::rrand(2, 10);
	decoration->size = 0.25;
	
	decoration->velocity.x /= 100;
	decoration->velocity.y /= 100;
	decoration->velocity.z /= 10;
	
	Math::addBit(&decoration->flags, EF_DYING);
}

void addDecoration(Properties *props)
{
	entityManager->spawnDecoration(DT_FLESH)->load(props);
}

void drawDecoration(Decoration *decoration)
{
	glPushMatrix();
	{
		switch (decoration->type)
		{
			case DT_SHARD:
				glEnable(GL_BLEND);
				glDisable(GL_TEXTURE_2D);
				glColor4f(0.85, 0.85, 1.0, decoration->custom);
				
				glTranslatef(decoration->position.x, decoration->position.y, decoration->position.z);
				
				glBegin(GL_TRIANGLES);
				{
					glVertex3f(decoration->triangle.point[0].x, decoration->triangle.point[0].y, decoration->triangle.point[0].z);
					glVertex3f(decoration->triangle.point[1].x, decoration->triangle.point[1].y, decoration->triangle.point[1].z);
					glVertex3f(decoration->triangle.point[2].x, decoration->triangle.point[2].y, decoration->triangle.point[2].z);
				}
				glEnd();
				break;
				
			default:
				/*
				(decoration->life > 50) ? glDisable(GL_BLEND) : glEnable(GL_BLEND);
				glEnable(GL_TEXTURE_2D);
			
				glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("Blood")->data);
				
				glTranslatef(decoration->position.x, decoration->position.y, decoration->position.z);
				
				glColor4fv(decoration->color.color);
				gluSphere(graphics->quadric, decoration->size, 16, 16);
				*/
				break;
		}
	}
	glPopMatrix();
}

void doDecoration()
{
	for (Decoration *decoration = (Decoration*)entityManager->decorationList.getFirstElement() ; decoration != NULL ; decoration = (Decoration*)decoration->next)
	{
		decoration->position += (decoration->velocity * engine->getTimeDifference(TD_LOGIC));
		
		decoration->thinkTime -= engine->getTimeDifference(TD_LOGIC);
		
		if (decoration->thinkTime <= 0)
		{
			switch (decoration->type)
			{
				case DT_FLESH:
					addBloodParticle(decoration->position.x, decoration->position.y, decoration->position.z);
					break;
				case DT_FIRE:
				case DT_EXPLOSION:
					addFireParticle(decoration->position.x, decoration->position.y, decoration->position.z);
					break;
			}
			
			decoration->thinkTime = Math::rrand(1, 5);
		}
		
		if (!(decoration->flags & EF_WEIGHTLESS))
		{
			decoration->velocity.z -= (0.01 * engine->getTimeDifference(TD_LOGIC));
		}
		
		//if (camera->containsSphere(decoration->position, 1))
		{
			drawDecoration(decoration);
		}
		
		if (decoration->flags & EF_DYING)
		{
			decoration->health -= engine->getTimeDifference(TD_LOGIC);
		}
		
		if (decoration->health <= -50)
		{
			decoration = (Decoration*)decoration->previous;
			entityManager->decorationList.remove(decoration->next);
		}
	}
}
