/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dialogs.h"

void moveMouseCursor()
{
	int x = (engine->axis1X / 10000);
	int y = (engine->axis1Y / 10000);
	
	if (game->isControl(CTRL_LEFT))
	{
		x = -5;
	}
	
	if (game->isControl(CTRL_RIGHT))
	{
		x = 5;
	}
	
	if (game->isControl(CTRL_UP))
	{
		y = -5;
	}
	
	if (game->isControl(CTRL_DOWN))
	{
		y = 5;
	}
	
	if ((x != 0) || (y != 0))
	{
		mouse->move(x, y);
	}
}

SDL_Surface *buildMainDialog(char *message)
{
	SDL_Surface *scratch = graphics->scratchSurface;
	SDL_FillRect(scratch, NULL, graphics->black);
	graphics->setTransparent(scratch);
	graphics->setFont(FONT_NORMAL);
	graphics->setFontColor(255, 255, 255, 0, 0, 0);
	graphics->writeWrappedText(scratch, true, message);
	
	int width = max(graphics->textWidth + 20, 250);

	SDL_Surface *dialogBox = graphics->createSurface(width, graphics->textHeight + 50);

	graphics->createRaisedBorder(dialogBox, graphics->lightGrey, graphics->grey);

	SDL_Rect source = {0, 0, graphics->textWidth, graphics->textHeight};
	SDL_Rect dest = {10, 10, graphics->textWidth, graphics->textHeight};

	if (SDL_BlitSurface(scratch, &source, dialogBox, &dest) < 0)
	{
		graphics->showErrorAndExit("dialogs::buildMainDialog - %s", SDL_GetError());
	}

	return dialogBox;
}

int showOKCancelDialog(char *message)
{
	glEnable(GL_BLEND);
	
	int rtn = -1;
	
	int x = graphics->screen->w / 2;
	int y = graphics->screen->h / 2;
	
	String text = localeManager->translate(message);
	Texture *dialogBox = graphics->createTexture(buildMainDialog(text.getText()), true);
	
	uiManager->OKCancelDialogGrid.x = (graphics->screen->w - 350) / 2;
	uiManager->OKCancelDialogGrid.width = 350;
	uiManager->OKCancelDialogGrid.height = 26;
	
	uiManager->OKCancelDialogGrid.y = y + (dialogBox->ih / 2) - 28;
	
	uiManager->OKCancelDialogGrid.updatePosition();
	
	Properties props;

	props.setProperty("groupName", "DialogBoxWidgets");
	props.setProperty("name", "tempDialogButton");
	props.setProperty("backgroundColor", "0 150 0");
	props.setProperty("enabled", "1");
	props.setProperty("visible", "1");
	props.setProperty("y", "0");
	props.setProperty("grid", "OKCancelDialogGrid");

	props.setProperty("label", "SYS_OK");
	props.setProperty("x", "1");
	Button *ok = (Button*)createButtonWidget(&props);

	props.setProperty("label", "GUI_CANCEL");
	props.setProperty("x", "2");
	Button *cancel = (Button*)createButtonWidget(&props);

	graphics->drawRect(x - (dialogBox->iw / 2), y - (dialogBox->ih / 2), dialogBox->iw, dialogBox->ih, GLColor::black, false);
	glColor4f(1.0, 1.0, 1.0, 1.0);
	glEnable(GL_BLEND);
	graphics->blit(dialogBox, x, y, true);
	
	uiManager->setModal("DialogBoxWidgets");

	while (rtn == -1)
	{
		engine->getInput();
		graphics->updateScreen();
		
		moveMouseCursor();

		drawWidgets();

		if (ok->wasClicked())
		{
			rtn = 1;
		}

		if (cancel->wasClicked())
		{
			rtn = 0;
		}
	}

	delete dialogBox;

	uiManager->setModal("");
	uiManager->removeGroup("DialogBoxWidgets");

	return rtn;
}

void showMessageDialog(char *message, char *buttonLabel)
{
	glEnable(GL_BLEND);
	
	debug(("showMessageDialog:: - %s\n", message));

	int x = graphics->screen->w / 2;
	int y = graphics->screen->h / 2;
	
	Texture *dialogBox = graphics->createTexture(buildMainDialog(message), true);
	
	Properties props;
	
	props.setProperty("x", (graphics->screen->w - 350) / 2);
	props.setProperty("y", y - (dialogBox->ih / 2) + dialogBox->ih - 40);
	props.setProperty("width", 350);
	props.setProperty("height", 26);
	uiManager->OKDialogGrid.setProperties(&props);
	
	props.clear();
	
	props.setProperty("groupName", "DialogBoxWidgets");
	props.setProperty("name", "tempDialogButton");
	props.setProperty("label", buttonLabel);
	props.setProperty("backgroundColor", "0 150 0");
	props.setProperty("enabled", "1");
	props.setProperty("visible", "1");
	props.setProperty("x", "1");
	props.setProperty("y", "0");
	props.setProperty("grid", "OKDialogGrid");
	
	Button *ok = (Button*)createButtonWidget(&props);

	graphics->setMode(MODE_2D);
	graphics->fade(0.85, x - (dialogBox->iw / 2), y - (dialogBox->ih / 2), dialogBox->iw, dialogBox->ih);
	glColor4f(1.0, 1.0, 1.0, 1.0);
	graphics->blit(dialogBox, x, y, true);
	
	uiManager->setModal("DialogBoxWidgets");
	
	mouse->setBusy(false);
	engine->doGameLogic = false;
	
	bool doneReading = false;

	while (!doneReading)
	{
		engine->getInput();
		graphics->updateScreen();
		graphics->clearScreen();
		
		engine->doTimeDifference();
		
		doGameObjects();
		
		graphics->setMode(MODE_2D);
		graphics->fade(0.85, x - (dialogBox->iw / 2), y - (dialogBox->ih / 2), dialogBox->iw, dialogBox->ih);
		glColor4f(1.0, 1.0, 1.0, 1.0);
		graphics->blit(dialogBox, x, y, true);
		
		moveMouseCursor();

		drawWidgets();
		
		if ((ok->wasClicked()) || (engine->userAccepts()) || (game->isControl(CTRL_ACTION)))
		{
			doneReading = true;
		}
	}
	
	engine->resetTimeDifference();
	
	mouse->setBusy(true);
	engine->doGameLogic = true;

	delete dialogBox;

	uiManager->setModal("");
	uiManager->removeGroup("DialogBoxWidgets");
	
	game->resetControl(CTRL_ACTION);
}

bool showSecurityDialog(char *requiredCode)
{
	debug(("showSecurityDialog() - Expecting '%s'\n", requiredCode));
	
	glEnable(GL_BLEND);
	
	int mx = graphics->screen->w / 2;
	int my = graphics->screen->h / 2;
	
	mx -= 96;
	my -= 96;
	
	engine->flushInput();
	
	int button;
	float buttonPressed[10], timeout;
	int px, py, key;
	String result;
	
	for (int i = 0 ; i < 10 ; i++)
	{
		buttonPressed[i] = 0;
	}
	
	timeout = 150;
	
	mouse->setBusy(false);
	engine->doGameLogic = false;
	
	doGameObjects();
	
	while (timeout > 0)
	{
		key = 0;
		
		engine->getInput();
		graphics->updateScreen();
		graphics->clearScreen();
		
		engine->doTimeDifference();
		
		doGameObjects();
		
		moveMouseCursor();
		
		graphics->setMode(MODE_2D);
		graphics->setFont(FONT_LARGE);
		
		if (result.getLength() < 4)
		{
			graphics->drawString(graphics->screen->w / 2, 160, TXT_CENTERED, GLColor::green, true, result.getText());
		}
		else
		{
			if (result == requiredCode)
			{
				graphics->drawString(graphics->screen->w / 2, 160, TXT_CENTERED, GLColor::green, true, "Code Accepted");
			}
			else
			{
				graphics->drawString(graphics->screen->w / 2, 160, TXT_CENTERED, GLColor::red, true, "Incorrect Code");
			}
		}
		
		for (int y = 0 ; y < 4 ; y++)
		{
			for (int x = 0 ; x < 3 ; x++)
			{
				px = mx + (x * 66);
				py = my + (y * 68);
				key++;
				button = key;
				
				if ((button == 10) || (button == 12))
				{
					continue;
				}
				else if (button == 11)
				{
					button = 0;
				}
				
				buttonPressed[button] -= engine->getTimeDifference(TD_ANIMATION);
				Math::limit(&buttonPressed[button], 0, 100);
				
				graphics->fade(0.95, px, py, 64, 64);
				graphics->outlineRect(px, py, 64, 64, (buttonPressed[button]) ? GLColor::red : GLColor::lightGrey);
				graphics->setFont(FONT_LARGE);
				graphics->drawString(px + 22, py + 14, TXT_LEFT, (buttonPressed[button]) ? GLColor::red : GLColor::lightGrey, true, "%d", button);
				
				
				if ((mouse->buttonState[SDL_BUTTON_LEFT]) && (result.getLength() < 4))
				{
					if (Collision::collision(mouse->x, mouse->y, 1, 1, px, py, 64, 64))
					{
						buttonPressed[button] = 10;
						mouse->buttonState[SDL_BUTTON_LEFT] = false;
						result.setText("%s%d", result.getText(), button);
					}
				}
			}
		}
		
		if (result.getLength() < 4)
		{
			key = (SDLK_0 - 1);
			
			while (key != SDLK_9)
			{
				key++;
				
				if (engine->keyState[key])
				{
					engine->keyState[key] = 0;
					
					result.setText("%s%d", result.getText(), (key - 48));
					
					buttonPressed[(key - 48)] = 10;
				}
			}
		}
		else if (result.getLength() == 4)
		{
			timeout -= engine->getTimeDifference(TD_ANIMATION);
		}
	}
	
	mouse->setBusy(true);
	engine->doGameLogic = true;
	
	return (result == requiredCode);
}

void populateSavesList(ListView *listView, bool loading)
{
	if (!loading)
	{
		addSaveHeaderToList(listView, NULL);
	}
	
	String saveDir;
	saveDir.setText("%s", engine->userHomeDirectory.getText());
	
	DIR *dir = opendir(saveDir.getText());
	FILE *fp;
	dirent *file;
	String saveData;
	
	SaveGameHeader *header;
	
	while ((file = readdir(dir)))
	{
		if (file->d_name[0] == '.')
		{
			continue;
		}
		
		saveData.setText("%s/%s/saveinfo", saveDir.getText(), file->d_name);
		
		fp = fopen(saveData.getText(), "rb");
		
		if (!fp)
		{
			printf("Ignoring '%s'\n", saveData.getText());
			continue;
		}
		
		fclose(fp);
		
		header = getSaveHeader(atoi(file->d_name));
		
		addSaveHeaderToList(listView, header);
	}
	
	closedir(dir);
}

int doLoadSaveGameDialog(char *type)
{
	uiManager->onlyShowGroup("LoadSave");
	
	ListView *savesList = (ListView*)uiManager->getWidget("SaveList", "LoadSave");
	Button *load = (Button*)uiManager->getWidget("Load", "LoadSave");
	Button *save = (Button*)uiManager->getWidget("Save", "LoadSave");
	Button *deleteGame = (Button*)uiManager->getWidget("Delete", "LoadSave");
	Button *cancel = (Button*)uiManager->getWidget("Cancel", "LoadSave");
	
	Grid *grid = uiManager->getGrid("LoadSaveGrid");
	
	uiManager->setFocusedWidget(savesList);
	
	load->setVisible(false);
	save->setVisible(false);
	uiManager->getWidget(type, "LoadSave")->setVisible(true);
	
	bool done = false;
	bool cancelled = false;
	
	populateSavesList(savesList, load->isVisible());
	
	ListViewItem *selectedItem = NULL;
	SaveGameHeader *header = NULL;
	
	if (savesList->getItemCount() > 0)
	{
		header = (SaveGameHeader*)savesList->getSelectedObject();
		deleteGame->setEnabled((header != NULL));
	}
	else
	{
		load->setEnabled(false);
		deleteGame->setEnabled(false);
	}
	
	/*
	if (game->saveSlot != -1)
	{
		if (strcmp(type, "Load") == 0)
		{
			printf("game->saveSlot = %d\n", game->saveSlot);
			savesList->setSelectedIndex(game->saveSlot);
		}
		else
		{
			printf("game->saveSlot = %d\n", game->saveSlot);
			savesList->setSelectedIndex(game->saveSlot + 1);
		}
	}
	*/
	
	int rtn = -1;
	
	mouse->setBusy(false);
	engine->doGameLogic = false;
	
	engine->resetTimeDifference();
	engine->clearInput();
	engine->flushInput();
	
	while (!done)
	{
		engine->doTimeDifference();
		
		graphics->updateScreen();
		engine->getInput();
		
		graphics->clearScreen();
		
		doGameObjects();
		
		graphics->setMode(MODE_2D);
		
		graphics->fade(0.5);
		
		graphics->drawRect(grid->x, grid->y, grid->width, grid->height, GLColor::black, GLColor::white, false);
		
		moveMouseCursor();
		
		drawWidgets();
		
		if (savesList->getItemCount() > 0)
		{
			if (savesList->itemSelectionChanged())
			{
				header = (SaveGameHeader*)savesList->getSelectedObject();
				selectedItem = savesList->getSelectedItem();
				deleteGame->setEnabled((header != NULL));
			}
		}
		
		if ((load->wasClicked()) || (save->wasClicked()) || (engine->keyState[SDLK_RETURN]))
		{
			break;
		}
		
		if (deleteGame->wasClicked())
		{
			//if (showOKCancelDialog("LOADSAVE_DELETE_CONFIRM"))
			{
				removeSaveFromList(header->slotId);
				delete header;
				savesList->removeItem(selectedItem);
				header = (SaveGameHeader*)savesList->getSelectedObject();
				selectedItem = savesList->getSelectedItem();
				deleteGame->setEnabled((header != NULL));
				
				if (savesList->getItemCount() > 0)
				{
					header = (SaveGameHeader*)savesList->getSelectedObject();
					deleteGame->setEnabled((header != NULL));
				}
				else
				{
					load->setEnabled(false);
					deleteGame->setEnabled(false);
				}
			}
		}
		
		if (cancel->wasClicked())
		{
			cancelled = true;
			done = true;
		}
	}
	
	if (header != NULL)
	{
		rtn = header->slotId;
	}
	
	// we must explictly destroy all these objects
	// because they are not stored anywhere else(!)
	List *l = savesList->getList();
	for (ListViewItem *item = (ListViewItem*)l->getFirstElement() ; item != NULL ; item = (ListViewItem*)item->next)
	{
		delete item->object;
	}
	
	savesList->getList()->clear();
	
	graphics->clearScreen();
	doGameObjects();
	
	mouse->setBusy(true);
	engine->doGameLogic = true;
	
	game->resetControl(CTRL_ACTION);
	
	return (!cancelled) ? rtn : -2;
}
