/*
Copyright (C) 2006 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "../headers.h"

Door::Door()
{
	entityType = ET_DOOR;
	
	lockCount = 0;
	
	initialLockCount = 0;
	
	state = DOOR_CLOSED;
	
	speed = 1;
	
	autoClose = false;
	
	autoCloseTimer = 0;
	
	name = "Door";
	
	textureSpec = NULL;
	model = NULL;
	
	flags = EF_NOCLIP;
}

Door::~Door()
{
}

void Door::loadAdditional(Properties *props)
{
	speed = props->getFloat("speed", speed);
	startPosition = props->getVector("startPosition", startPosition);
	endPosition = props->getVector("endPosition", endPosition);
	
	lockCount = props->getInt("lockCount", lockCount);
	
	if (props->hasProperty("initialLockCount"))
	{
		initialLockCount = props->getInt("initialLockCount", initialLockCount);
	}
	else
	{
		initialLockCount = lockCount;
	}
	
	autoClose = props->getInt("autoClose", false);
	autoCloseTimer = props->getFloat("autoCloseTimer", autoCloseTimer);
	
	MDLModelName = props->getString("model", MDLModelName.getText());
}

void Door::saveAdditional(Properties *props)
{
	props->setProperty("state", state);	
	props->setProperty("speed", speed);
	props->setProperty("startPosition", startPosition.toString());
	props->setProperty("endPosition", endPosition.toString());
	
	props->setProperty("lockCount", lockCount);
	props->setProperty("initialLockCount", initialLockCount);
	props->setProperty("autoClose", autoClose);
	props->setProperty("autoCloseTimer", autoCloseTimer);
	
	props->setProperty("model", MDLModelName.getText());
}
