/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "items.h"

void throwSpawnedItem(Entity *item)
{
	item->velocity.x = Math::rrand(-2, 2);
	item->velocity.y = Math::rrand(-2, 2);
	item->velocity.z = Math::rrand(5, 10);
	
	item->velocity.x /= 10;
	item->velocity.y /= 10;
	item->velocity.z /= 10;
}

int handleMedicalItem(Unit *unit, Item *item, bool pickupNow)
{
	if ((unit->health == unit->maxHealth) && (!(item->flags & EF_AUTO_PICKUP)))
	{
		return ITEM_LEAVE;
	}
	
	if (!pickupNow)
	{
		return ITEM_PICKUP;
	}
				
	if (unit == player)
	{
		game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), ((ItemDef*)item->definition)->getName());
		game->alterStat("CHERRIES_PICKED_UP", 1);
	}
				
	audio->playSound(SND_GULP, CH_ANY);
				
	unit->health += item->itemPower;
	Math::limit(&unit->health, 0, unit->maxHealth);
	
	return ITEM_CONSUME;
}

int handlePowerUp(Item *item, bool pickupNow)
{
	if (!pickupNow)
	{
		return ITEM_PICKUP;
	}
	
	WeaponDef *weaponDef;
	ItemDef *itemDef = (ItemDef*)item->definition;
	Unit *teeka;
	int weaponType;
	
	String description;
	
	switch (itemDef->type)
	{
		case ITM_SHIELD_BELT:
			player->maxShield = max(player->maxShield, item->itemPower);
			game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), itemDef->getName());
			addPowerUpParticles(player->position);
			break;
			
		case ITM_TEEKA_WEAPON:
			weaponType = engine->getValueOfDefine(item->getName());
			weaponDef = &entityManager->weaponDef[weaponType];
			
			game->teekaWeapon = weaponType;
			
			teeka = findTeeka();
			teeka->currentWeapon = weaponDef;
			addPowerUpParticles(teeka->position);
			
			description.setText(itemDef->getName(), weaponDef->getName());
			game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), description.getText());
			break;
	}
	
	//audio->playSound(SND_POWERUP, CH_ANY);
	audio->playSound(SND_ITEM, CH_ANY);
	
	game->alterStat("ITEMS_PICKED_UP", 1);
	
	return ITEM_CONSUME;
}

int handleSpecialItem(Item *item, bool pickupNow)
{
	int freeSlot;
	
	if (((freeSlot = game->getFreeInventorySlot()) == -1) && (!(item->flags & EF_AUTO_PICKUP)))
	{
		return ITEM_LEAVE;
	}
	
	// Player must always pick these up themselves...
	if (!pickupNow)
	{
		return ITEM_LEAVE;
	}
	
	item->setOwner(player);
	
	game->item[freeSlot] = item;
	
	game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), ((ItemDef*)item->definition)->getName());
	
	game->alterStat("ITEMS_PICKED_UP", 1);
	
	audio->playSound(SND_ITEM, CH_ANY);
	
	return ITEM_PICKUP;
}

int handleWeaponItem(Weapon *weapon, bool pickupNow)
{
	int freeSlot = -1;
	int currentWeaponSlot = -1;
	
	WeaponDef *weaponDef = (WeaponDef*)weapon->definition;
		
	for (int i = 3 ; i > -1 ; i--)
	{
		if (game->weapon[i] == NULL)
		{
			freeSlot = i;
		}
		else if (game->weapon[i]->definition == weapon->definition)
		{
			currentWeaponSlot = i;
		}
	}
		
	// Does not have the weapon and has a free slot
	if ((currentWeaponSlot == -1) && (freeSlot != -1))
	{
		if (!pickupNow)
		{
			return ITEM_PICKUP;
		}
		
		weapon->setOwner(player);
		
		game->weapon[freeSlot] = weapon;
		game->weapon[freeSlot]->currentAmmo = weapon->currentAmmo;
		game->selectedWeapon = game->weapon[freeSlot];
		
		game->weaponIndex = freeSlot;
			
		audio->playSound(SND_PICKUP_WEAPON, CH_PICKUP);
		game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), weaponDef->getName());
		game->alterStat("WEAPONS_PICKED_UP", 1);
			
		return ITEM_PICKUP;
	}
	else if (currentWeaponSlot != -1)
	{
		if (game->weapon[currentWeaponSlot]->currentAmmo ==	weaponDef->clipSize)
		{
			if (weapon->flags & EF_AUTO_PICKUP)
			{
				audio->playSound(SND_PICKUP_WEAPON, CH_PICKUP);
				return ITEM_CONSUME;
			}
			
			return ITEM_LEAVE;
		}
		
		if (!pickupNow)
		{
			return ITEM_PICKUP;
		}
			
		// auto switch to better weapon
		if ((game->selectedWeapon == game->weapon[0]) && (game->weapon[currentWeaponSlot]->currentAmmo == 0))
		{
			game->selectedWeapon = game->weapon[currentWeaponSlot];
			game->weaponIndex = currentWeaponSlot;
		}
			
		game->weapon[currentWeaponSlot]->currentAmmo += weapon->currentAmmo;
			
		Math::limit(&game->weapon[currentWeaponSlot]->currentAmmo, 0, weaponDef->clipSize);
			
		audio->playSound(SND_PICKUP_WEAPON, CH_PICKUP);
		
		if (weaponDef->type != WEP_GRENADE)
		{
			game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_GOT_AMMO"), weaponDef->getName());
		}
		else
		{
			game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_GOT_GRENADES"), weaponDef->getName());
		}
		
		game->alterStat("WEAPONS_PICKED_UP", 1);
		
		textureManager->removeTexture("str_current_ammo");
		String s;
		s.setText("str_weapon_%d", currentWeaponSlot);
		textureManager->removeTexture(s.getText());
			
		return ITEM_CONSUME;
	}
	else if ((freeSlot == -1) && (game->selectedWeapon != game->weapon[0]) && (game->selectedWeapon != game->weapon[3]))
	{
		game->setGamePlayMessage(localeManager->translate("HINT_SWAP_WEAPON"), game->selectedWeapon->definition->getName(), weaponDef->getName());
			
		if (game->isControl(CTRL_ACTION))
		{
			game->resetControl(CTRL_ACTION);
				
			int i = game->getSelectedWeaponSlot();
				
			// put first weapon back on map...
			entityManager->weaponList.add(game->weapon[i]);
			game->weapon[i]->position = player->position;
			game->weapon[i]->custom = 100;
			game->weapon[i]->setOwner(NULL);
			throwSpawnedItem(game->weapon[i]);
				
			// get second weapon
			entityManager->weaponList.detach(weapon);
			game->weapon[i] = weapon;
			game->weapon[i]->currentAmmo = weapon->currentAmmo;
			game->weapon[i]->setOwner(player);
			game->selectedWeapon = game->weapon[i];
			
			audio->playSound(SND_PICKUP_WEAPON, CH_PICKUP);
			game->setInfoMessage(INFO_PICKUP, localeManager->translate("INFO_PICKED_UP"), weaponDef->getName());
				
			return ITEM_PICKUP;
		}
	}
	
	return ITEM_LEAVE;
}

int pickUpItem(Unit *unit, Entity *ent, bool pickupNow)
{
	if (ent->entityType == ET_ITEM)
	{
		Item *item = (Item*)ent;
		ItemDef *itemDef = (ItemDef*)item->definition;
		
		switch (itemDef->itemType)
		{
			case IT_MEDICAL:
				return handleMedicalItem(unit, item, pickupNow);
				break;
				
			case IT_POWER_UP:
				if (unit == player)
				{
					return handlePowerUp(item, pickupNow);
				}
				return ITEM_LEAVE;
				break;
				
			case IT_SPECIAL:
				if (unit == player)
				{
					return handleSpecialItem(item, pickupNow);
				}
				return ITEM_LEAVE;
				break;
		}
	}
	else if (unit == player)
	{
		return handleWeaponItem((Weapon*)ent, pickupNow);
	}
	
	return ITEM_LEAVE;
}

void pickupItem(Entity *other)
{
	if (other == NULL)
	{
		return;
	}
	
	if ((self->custom > 0) || (other->health <= 0) || (other->entityType != ET_BLOB))
	{
		return;
	}
	
	int result = pickUpItem((Unit*)other, self, true);
	
	if (result != ITEM_LEAVE)
	{
		if (self->name != "")
		{
			fireTriggers(self->getName(), TRIGGER_TYPE_ITEM_PICKED_UP);
		}
		
		debug(("%s picked up %s\n", other->getName(), self->definition->getName()));
		
		// picked up an item that can be removed (eg - cherry or ammo)
		if (result == ITEM_CONSUME)
		{
			self->health = 0;
		}
		else
		{
			if (self->entityType == ET_WEAPON)
			{
				entityManager->weaponList.detach(self);
			}
			else if (self->entityType == ET_ITEM)
			{
				entityManager->itemList.detach(self);
			}
		}
	}
	
	// tell the unit to immediately stop the item hunting
	if ((other != player) && (result == ITEM_CONSUME))
	{
		other->action = other->walk;
		other->custom = 0;
		other->velocity.x = other->velocity.y = 0;
		other->thinkTime = 0;
	}
}

Item *addItem(Properties *props)
{
	int type = entityManager->getEntityType(props);
	
	if (type == -1)
	{
		printf("ERROR: addItem() - Type not known!!\n");
		exit(1);
	}
	
	Item *item = entityManager->spawnItem(type);
	item->load(props);
	
	if (!game->allStatic)
	{
		item->touch = &pickupItem;
	}
	
	debug(("addItem() - %s [%d, %s]\n", item->definition->getName(), item->getUniqueId(), item->position.toString()));
	
	return item;
}

Weapon *addWeapon(Properties *props)
{
	int type = entityManager->getEntityType(props);
	
	if (type == -1)
	{
		printf("ERROR: addWeapon() - Type not known!!\n");
		exit(1);
	}
	
	Weapon *weapon = entityManager->spawnWeapon(type, false);
	weapon->load(props);
	
	if (!game->allStatic)
	{
		weapon->touch = &pickupItem;
	}
	
	debug(("addWeapon() - %s [%d, %s]\n", weapon->definition->getName(), weapon->getUniqueId(), weapon->position.toString()));
	
	return weapon;
}

void bringItemToPlayer(Entity *item)
{
	/*
	 * First check that we're allowed to pick up this item.
	 * If not then just leave it there.
	*/
	if (!(item->flags & EF_AUTO_PICKUP))
	{
		if (pickUpItem(player, item, false) == ITEM_LEAVE)
		{
			return;
		}
	}
	
	if (player->helpless)
	{
		return;
	}
	
	item->velocity = player->position;
	item->velocity -= item->position;
	item->velocity.normalize();
	item->velocity = item->velocity * 2;
	item->thinkTime = Math::rrand(1, 3);
}

void processItemList(List *list)
{
	for (Entity *item = (Entity*)list->getFirstElement() ; item != NULL ; item = (Entity*)item->next)
	{
		self = item;
		
		bbManager->removeBox(item);
		
		Math::limit(&(--item->custom), 0, RAND_MAX);
		Math::limit(&(item->thinkTime -= engine->getTimeDifference(TD_LOGIC)), 0, RAND_MAX);
		
		if (withinViewableRange(item->position, 1))
		{
			//drawBoundingBox(item);
			
			glPushMatrix();
			{
				glColor3f(1.0, 1.0, 1.0);
				
				glTranslatef(item->position.x, item->position.y, item->position.z);
				
				glRotatef(90.0, 1.0, 0.0, 0.0);
				
				glRotatef(item->rotation.y, 0.0, 1.0, 0.0);
				
				item->rotation.y += engine->getTimeDifference(TD_ANIMATION);
				
				glBindTexture(GL_TEXTURE_2D, item->definition->texture->data);
			
				glCallList(item->definition->model->data);
			}
			glPopMatrix();
		}
		
		item->applyGravity(engine->getTimeDifference(TD_LOGIC));
		
		moveEntity(item);
		
		if ((item->flags & EF_ONGROUND) && (!(item->flags & EF_BOUNCES)))
		{
			item->velocity.z = 0;
		}
		
		item->contentType = tracer->outputContent;
		
		if (item->flags & EF_ONGROUND)
		{
			item->velocity.x *= 0.95;
			item->velocity.y *= 0.95;
		}
		
		if (item->contentType & CONTENTS_WATER)
		{
			Math::removeBit(&item->flags, EF_BOUNCES);
			Math::limit(&item->velocity.z, -0.25, 0.25);
		}
		else if ((item->contentType & CONTENTS_LAVA) || (item->contentType & CONTENTS_SLIME))
		{
			Math::addBit(&item->flags, EF_AUTO_PICKUP);
			Math::addBit(&item->flags, EF_WEIGHTLESS);
			Math::addBit(&item->flags, EF_NOCLIP);
			item->thinkTime = 0;
		}
		
		if (!game->allStatic)
		{
			if ((item->flags & EF_AUTO_PICKUP) && (item->thinkTime <= 0) && (game->missionStatus == MISSION_INPROGRESS))
			{
				bringItemToPlayer(item);
			}
			
			if (Math::getDistance(player->position, item->position) < 25)
			{
				bringItemToPlayer(item);
			}
		}
		
		if (item->health <= 0)
		{
			debug(("Removing Item '%s'\n", item->definition->getName()));
			item = (Entity*)item->previous;
			list->remove(item->next);
			continue;
		}
		
		bbManager->addBox(item);
	}
}

void doItems()
{
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	
	processItemList(&entityManager->weaponList);
	processItemList(&entityManager->itemList);
}
