/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "load.h"

void assignItemToUnit(Item *item)
{
	Unit *unit = entityManager->getOwnerById(item->ownerId);
	
	item->setOwner(unit);
	unit->currentItem = item;
	
	debug(("assignItemToUnit() - '%s' to '%s'\n", item->getName(), unit->getName()));
}

void reassignItems()
{
	debug(("reassignItems()\n"));
	
	Entity *entity, *next;
	
	debug(("reassignItems() - Weapons: %d\n", entityManager->weaponList.getSize()));
	debug(("reassignItems() - Items: %d\n", entityManager->itemList.getSize()));
	debug(("reassignItems() - Enemies: %d\n", entityManager->enemyList.getSize()));
	
	for (entity = (Entity*)entityManager->weaponList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		if ((entity->ownerId != 0) && (entity->ownerId != player->getUniqueId()))
		{
			printf("ERROR: Weapons can only be carried by the player!!\n");
			exit(1);
		}
	}
	
	for (entity = (Entity*)entityManager->itemList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		next = entity;
		
		if ((entity->ownerId != 0) && (entity->ownerId != player->getUniqueId()))
		{
			entity = (Entity*)entity->previous;
			entityManager->itemList.detach(next);
			assignItemToUnit((Item*)next);
		}
	}
	
	if (game->changingLevel)
	{
		return;
	}
	
	for (int i = 0 ; i < 4 ; i++)
	{
		if (game->weaponId[i] != 0)
		{
			debug(("reassignItems() - Getting weapon %d: %d\n", i, game->weaponId[i]));
			
			game->weapon[i] = entityManager->getWeaponById(game->weaponId[i]);
			game->weapon[i]->setOwner(player);
			
			entityManager->weaponList.detach(game->weapon[i]);
		}
	}
	
	for (int i = 0 ; i < 5 ; i++)
	{
		if (game->itemId[i] != 0)
		{
			debug(("reassignItems() - Getting item %d: %d\n", i, game->itemId[i]));
			
			game->item[i] = entityManager->getItemById(game->itemId[i]);
			game->item[i]->setOwner(player);
			
			entityManager->itemList.detach(game->item[i]);
		}
	}
	
	game->selectedWeapon = game->weapon[game->weaponIndex];
	
	debug(("reassignItems() :: Done\n"));
}

void reassignBullets()
{
	debug(("reassignBullets()\n"));
	
	for (Entity *entity = (Entity*)entityManager->bulletList.getFirstElement() ; entity != NULL ; entity = (Entity*)entity->next)
	{
		entity->setOwner(entityManager->getOwnerById(entity->ownerId));
		
		debug(("Bullet owner = %s (%d)\n", entity->owner->getName(), entity->owner->getUniqueId()));
	}
	
	debug(("reassignBullets() :: Done\n"));
}

void reassignOwnerShips()
{
	reassignItems();
	reassignBullets();
}

void decompressSaveData(int slotId, char *mission)
{
	debug(("decompressSaveData()\n"));
	
	String filename;
	
	filename.setText("%s/"SAVE_FORMAT"/%s", engine->userHomeDirectory.getText(), slotId, mission);
	
	FILE *fp = fopen(filename.getText(), "rb");
	
	if (!fp)
	{
		graphics->showErrorAndExit("Save game file '%s' is unavailable", filename.getText());
	}
	
	uLongf cSize, fSize;
	
	fseek(fp, (-sizeof(uLongf)) * 2, SEEK_END);
	fread(&cSize, sizeof(uLongf), 1, fp);
	fread(&fSize, sizeof(uLongf), 1, fp);
	
	debug(("decompressSaveData() - cSize: %ld\n", cSize));
	debug(("decompressSaveData() - fSize: %ld\n", fSize));
	
	unsigned char *input = NULL;
	
	if (engine->dataBuffer != NULL)
	{
		delete[] engine->dataBuffer;
	}

	engine->dataBuffer = NULL;

	input = new unsigned char[(int)(cSize * 1.01) + 12];
	engine->dataBuffer = new unsigned char[fSize];
	
	debug(("decompressSaveData() - Opening '%s'\n", filename.getText()));

	rewind(fp);
	fread(input, 1, cSize, fp);

	fclose(fp);
	
	uncompress(engine->dataBuffer, &fSize, input, cSize);
	
	if (input != NULL)
	{
		delete[] input;
	}
	
	input = NULL;
	
	debug(("decompressSaveData() :: Done\n"));
}

SaveGameHeader *getSaveHeader(int slotId)
{
	char date[256], time[256], mission[256], description[1024];
	
	String fileDir;
	fileDir.setText(SAVE_FORMAT, slotId);
	
	String saveData;
	saveData.setText("%s/%s/saveinfo", engine->userHomeDirectory.getText(), fileDir.getText());
		
	FILE *fp = fopen(saveData.getText(), "rb");
		
	fscanf(fp, "%[^;]%*c%*c", date);
	fscanf(fp, "%[^;]%*c%*c", time);
	fscanf(fp, "%[^;]%*c%*c", description);
	fscanf(fp, "%[^;]%*c%*c", mission);
	
	fclose(fp);
		
	SaveGameHeader *header = new SaveGameHeader();
	
	header->slotId = slotId;
	header->date = date;
	header->time = time;
	header->description = description;
	header->mission = mission;
	
	return header;
}

bool loadGameData(int slotId, char *mission)
{
	debug(("loadGameData() - %s\n", mission));
	
	engine->startDebugTimer("loadGameData()");
	
	if (!game->changingLevel)
	{
		game->clear();
		entityManager->clear();
	}
	
	decompressSaveData(slotId, mission);
	
	List *list = loadPropertiesInMemory();
	
	if (list == NULL)
	{
		return false;
	}
	
	graphics->showLoadingProgress("Loading Game...");
	
	setupEntities(list);
	
	reassignOwnerShips();
	
	engine->endDebugTimer("loadGameData()");
	
	removeBSPModelFaces();

	return true;
}

void loadGameData(int slot)
{
	debug(("loadGameData() - %d\n", slot));
	
	SaveGameHeader *header = getSaveHeader(slot);
	
	loadGameData(slot, header->mission.getText());
	
	delete header;
}

bool loadGame()
{
	int saveSlot = doLoadSaveGameDialog("Load");
	
	// don't bother! They pressed cancel!
	if (saveSlot == -2)
	{
		return false;
	}
	
	game->saveSlot = saveSlot;
	
	uiManager->onlyShowGroup("@none@");
	
	engine->resetTimeDifference();
	
	return true;
}
