/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USAf.

*/

#include "player.h"

void blobWalkTurn()
{
}

void resetPlayer()
{
	player->helpless = 0;
	player->position = game->checkpoint;
	player->velocity.set(0, 0, 0);
	player->rotation.set(0, 0, 0);
	player->contentType = CONTENTS_SOLID;
	Math::removeBit(&player->flags, EF_BOUNCES);
	game->setMissionStatus(MISSION_INPROGRESS);
}

void calculatePlayerAimDirection()
{
	game->targetLock = player->position;
	
	float r = toRadians(player->rotation.x);
	
	game->targetLock.x += (sin(r) * 25);
	game->targetLock.y -= (cos(r) * 25);
	
	Vector v = game->targetLock;
	v -= player->position;
	v.normalize();
	
	game->targetLock += (v * 150);
	
	//game->targetLock = graphics->convertToWorldCoords(game->aimDirection.x, game->aimDirection.y, NULL, NULL, NULL);
}

void drawLaserSight(Unit *unit)
{
	/*
	if (unit->target == NULL)
	{
		return;
	}
	*/
	
	glPushMatrix();
	{
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		
		glBegin(GL_LINES);
		{
			glColor4f(1.0, 0.0, 0.0, 0.5); glVertex3f(player->position.x, player->position.y, player->position.z + 1.5);
			glColor4f(1.0, 0.0, 0.0, 0.25); glVertex3f(game->targetLock.x, game->targetLock.y, game->targetLock.z + 1.5);
		}
		glEnd();
		
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
	}
	glPopMatrix();
}

Entity *playerFavourMovingTarget(Unit *wanted)
{
	if (player->target == NULL)
	{
		return wanted;
	}
	
	// assume wanted target is closest
	if (player->target->flags & EF_STATIC)
	{
		return wanted;
	}
	else if (wanted->flags & EF_STATIC)
	{
		return player->target;
	}
	
	return wanted;
}

void playerTargetStructure()
{
	float distance = 0;
	float targetDistance = 226;
	
	Vector sight;
	
	for (Structure *structure = (Structure*)entityManager->elementList.getFirstElement() ; structure != NULL ; structure = (Structure*)structure->next)
	{
		if ((structure->flags & EF_IMMORTAL) || (structure->health <= 0) || (!camera->containsSphere(structure->position, 6)))
		{
			continue;
		}
		
		if (structure->entityType == ET_WINDOW)
		{
			continue;
		}
		
		distance = Math::getDistance(player->position, structure->position);
		
		if (distance < targetDistance)
		{
			sight = structure->position;
			sight.z += 6;
			
			if (unitCanSeeTarget(player, structure, sight))
			{
				targetDistance = distance;
				player->target = structure;
			}
		}
	}
}

void playerTargetEnemy()
{
	player->target = NULL;
	
	float distance = 0;
	float targetDistance = 226;
	
	Vector sight;
	
	for (Unit *unit = (Unit*)entityManager->enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if ((unit->health <= 0) || (!camera->containsSphere(unit->position, 6)) || (unit->flags & EF_NO_TARGET))
		{
			continue;
		}
		
		distance = Math::getDistance(player->position, unit->position);
		
		if (distance < targetDistance)
		{
			sight = unit->position;
			sight.z += 6;
			
			if (unitCanSeeTarget(player, unit, sight))
			{
				targetDistance = distance;
				player->target = playerFavourMovingTarget(unit);
			}
		}
	}
	
	if (player->target == NULL)
	{
		playerTargetStructure();
	}
}

void playerCycleStructures()
{
	float distance = 0;
	float targetDistance = 226;
	float currentDistance = Math::getDistance(player->position, player->target->position);
	Entity *oldTarget = player->target;
	Vector sight;
	
	for (Structure *structure = (Structure*)entityManager->elementList.getFirstElement() ; structure != NULL ; structure = (Structure*)structure->next)
	{
		if ((structure->flags & EF_IMMORTAL) || (structure->health <= 0) || (!camera->containsSphere(structure->position, 6)))
		{
			continue;
		}
		
		distance = Math::getDistance(player->position, structure->position);
		
		if ((distance < targetDistance) && (distance > currentDistance))
		{
			sight = structure->position;
			
			if (unitCanSeeTarget(player, structure, sight))
			{
				targetDistance = distance;
				player->target = structure;
			}
		}
	}
	
	if (oldTarget == player->target)
	{
		playerTargetStructure();
	}
}

void playerCycleTargets()
{
	float distance = 0;
	float targetDistance = 226;
	float currentDistance = Math::getDistance(player->position, player->target->position);
	Entity *oldTarget = player->target;
	Vector sight;
	
	bool cycleStructures = (oldTarget->entityType != ET_BIOMECH);
	
	for (Unit *unit = (Unit*)entityManager->enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if ((unit->health <= 0) || (!camera->containsSphere(unit->position, 6)) || (unit == player->target))
		{
			continue;
		}
		
		distance = Math::getDistance(player->position, unit->position);
		
		if ((distance < targetDistance) && (distance > currentDistance))
		{
			sight = unit->position;
			
			if (unitCanSeeTarget(player, unit, sight))
			{
				targetDistance = distance;
				player->target = playerFavourMovingTarget(unit);
				
				cycleStructures = false;
			}
		}
	}
	
	if (cycleStructures)
	{
		playerCycleStructures();
	}
	else if (oldTarget == player->target)
	{
		playerTargetEnemy();
	}
}

void playerSelectBestWeapon()
{
	game->mustChangeWeapon = false;
	
	if ((game->weapon[1] != NULL) && (game->weapon[1]->currentAmmo > 0))
	{
		game->selectedWeapon = game->weapon[1];
		game->weaponIndex = 1;
		player->reload = 100;
		return;
	}
	else if ((game->weapon[2] != NULL)&& (game->weapon[2]->currentAmmo > 0))
	{
		game->selectedWeapon = game->weapon[2];
		game->weaponIndex = 2;
		player->reload = 100;
		return;
	}
	
	game->selectedWeapon = game->weapon[0];
	game->weaponIndex = 0;
	player->reload = 100;
}

void playerIssueOrders()
{
	if (engine->keyState[SDLK_7])
	{
		game->setInfoMessage(INFO_HINT, "Orders : Follow");
		game->currentOrders = ORDERS_FOLLOW;
		engine->keyState[SDLK_7] = 0;
	}
	
	if (engine->keyState[SDLK_8])
	{
		game->setInfoMessage(INFO_HINT, "Orders : Wait");
		game->currentOrders = ORDERS_WAIT;
		engine->keyState[SDLK_8] = 0;
	}
	
	if (engine->keyState[SDLK_9])
	{
		game->setInfoMessage(INFO_HINT, "Orders : Use");
		game->currentOrders = ORDERS_USE;
		engine->keyState[SDLK_9] = 0;
		resetBlobOrders();
	}
	
	if (engine->keyState[SDLK_0])
	{
		game->setInfoMessage(INFO_HINT, "Orders : Retreat");
		game->currentOrders = ORDERS_RETREAT;
		engine->keyState[SDLK_0] = 0;
	}
}

void playerFireWeapon(Weapon *weapon)
{
	if (weapon->currentAmmo > 0)
	{
		// can't fire underwater :)
		if ((player->liquidLevel > 1) && ((WeaponDef*)weapon->definition->type != WEP_PISTOL))
		{
			return;
		}
		
		if (player->target != NULL)
		{
			fireWeapon(player, (WeaponDef*)weapon->definition, player->position, player->target->position);
		}
		else
		{
			calculatePlayerAimDirection();
			//game->targetLock.z = player->position.z;
			fireWeapon(player, (WeaponDef*)weapon->definition, player->position, game->targetLock);
		}
			
		if (!(weapon->definition->flags & WF_UNLIMITED))
		{
			weapon->currentAmmo--;
			textureManager->removeTexture("str_current_ammo");
			
			String s;
			s.setText("str_weapon_%d", game->weaponIndex);
			textureManager->removeTexture(s.getText());
				
			if (weapon->currentAmmo == 0)
			{
				game->mustChangeWeapon = true;
			}
		}
	}
	else
	{
		audio->playSound(SND_WEAPON_EMPTY, CH_PLAYER);
		player->reload = 8;
	}
}

void doPlayerCycleWeapon(int dir)
{
	Math::wrap(&(game->weaponIndex += dir), 0, 3);
		
	while (game->weapon[game->weaponIndex] == NULL)
	{
		Math::wrap(&(game->weaponIndex += dir), 0, 3);
	}
		
	game->selectedWeapon = game->weapon[game->weaponIndex];
		
	game->resetControl(CTRL_CYCLE_WEAPON);
		
	textureManager->removeTexture("str_current_ammo");
}

void doPlayerWeapons()
{
	if (game->isControl(CTRL_LOCK_TARGET))
	{
		if (player->target == NULL)
		{
			playerTargetEnemy();
		}
		else
		{
			player->target = NULL;
		}
		
		game->resetControl(CTRL_LOCK_TARGET);
	}
	
	if (game->isControl(CTRL_CYCLE_TARGET))
	{
		if (player->target == NULL)
		{
			playerTargetEnemy();
		}
		else
		{
			playerCycleTargets();
		}
		
		game->resetControl(CTRL_CYCLE_TARGET);
	}
	
	/*
	if (game->isControl(CTRL_CYCLE_TARGET))
	{
		game->strafeLock = !game->strafeLock;
		game->aimDirection.set(400, 200);
		game->resetControl(CTRL_CYCLE_TARGET);
	}
	*/
	
	if (game->isControl(CTRL_CYCLE_WEAPON))
	{
		doPlayerCycleWeapon(1);
	}
	
	if (player->target != NULL)
	{
		faceTarget(player);
	}
	
	if (engine->keyState[SDLK_1])
	{
		game->weaponIndex = 0;
		game->selectedWeapon = game->weapon[game->weaponIndex];
		engine->keyState[SDLK_1] = 0;
	}
	
	if (engine->keyState[SDLK_2])
	{
		if (game->weapon[1] != NULL)
		{
			game->weaponIndex = 1;
			game->selectedWeapon = game->weapon[1];
		}
		engine->keyState[SDLK_2] = 0;
	}
	
	if (engine->keyState[SDLK_3])
	{
		if (game->weapon[2] != NULL)
		{
			game->weaponIndex = 2;
			game->selectedWeapon = game->weapon[2];
		}
		engine->keyState[SDLK_3] = 0;
	}
	
	if (engine->keyState[SDLK_4])
	{
		game->weaponIndex = 3;
		
		if (game->weapon[game->weaponIndex]->currentAmmo > 0)
		{
			game->selectedWeapon = game->weapon[3];
		}
		engine->keyState[SDLK_4] = 0;
	}
	
	if (mouse->buttonState[SDL_BUTTON_WHEELUP])
	{
		doPlayerCycleWeapon(-1);
		mouse->buttonState[SDL_BUTTON_WHEELUP] = 0;
	}
	
	if (mouse->buttonState[SDL_BUTTON_WHEELDOWN])
	{
		doPlayerCycleWeapon(1);
		mouse->buttonState[SDL_BUTTON_WHEELDOWN] = 0;
	}
	
	if ((game->isControl(CTRL_FIRE)) && (player->reload == 0))
	{
		playerFireWeapon(game->selectedWeapon);
	}
	else if ((game->isControl(CTRL_QUICK_GRENADE)) && (player->reload == 0))
	{
		playerFireWeapon(game->weapon[3]);
		textureManager->removeTexture("str_weapon_3");
	}
	
	if ((game->mustChangeWeapon) && (!game->isControl(CTRL_FIRE)))
	{
		playerSelectBestWeapon();
	}
}

void movePlayerMouse()
{
	int x, y;
	
	SDL_GetRelativeMouseState(&x, &y);
	
	float turnSpeed = -x;
	turnSpeed /= game->mouseSensitivity;
	
	game->changeCameraRotation(turnSpeed * engine->getTimeDifference(TD_LOGIC));
}

void movePlayerAnalog()
{
	if (engine->joystick == NULL)
	{
		return;
	}
	
	float speed;
	
	if (engine->axis1X <= -game->joystickDeadZone)
	{
		speed = engine->axis1X;
		speed /= ANALOG_MOVE;
		
		player->velocity.x += (PLAYER_ACCEL * speed * -sin(game->cameraAdjustedRadian - 1.5));
		player->velocity.y += (PLAYER_ACCEL * speed * cos(game->cameraAdjustedRadian - 1.5));
		player->rotation.x = 270 + toDegrees(game->cameraAdjustedRadian);
	}
	
	if (engine->axis1X >= game->joystickDeadZone)
	{
		speed = engine->axis1X;
		speed /= ANALOG_MOVE;
		
		player->velocity.x += (PLAYER_ACCEL * speed * -sin(game->cameraAdjustedRadian - 1.5));
		player->velocity.y += (PLAYER_ACCEL * speed * cos(game->cameraAdjustedRadian - 1.5));
		player->rotation.x = 90 + toDegrees(game->cameraAdjustedRadian);
	}
	
	if (engine->axis1Y >= game->joystickDeadZone)
	{
		speed = ~engine->axis1Y;
		speed /= ANALOG_MOVE;
		
		player->velocity.x += (PLAYER_ACCEL * speed * -sin(game->cameraAdjustedRadian));
		player->velocity.y += (PLAYER_ACCEL * speed * cos(game->cameraAdjustedRadian));
		player->rotation.x = toDegrees(game->cameraAdjustedRadian);
		
		if (engine->axis1X <= -game->joystickDeadZone)
		{
			player->rotation.x -= 45;
		}
			
		if (engine->axis1X >= game->joystickDeadZone)
		{
			player->rotation.x += 45;
		}
	}
	
	if (engine->axis1Y <= -game->joystickDeadZone)
	{
		speed = ~engine->axis1Y;
		speed /= ANALOG_MOVE;
		
		player->velocity.x += (PLAYER_ACCEL * speed * -sin(game->cameraAdjustedRadian));
		player->velocity.y += (PLAYER_ACCEL * speed * cos(game->cameraAdjustedRadian));
		player->rotation.x = 180 + toDegrees(game->cameraAdjustedRadian);
		
		if (engine->axis1X <= -game->joystickDeadZone)
		{
			player->rotation.x += 45;
		}
			
		if (engine->axis1X >= game->joystickDeadZone)
		{
			player->rotation.x -= 45;
		}
	}
	
	if ((engine->axis2X <= -game->joystickDeadZone) || (engine->axis2X >= game->joystickDeadZone))
	{
		float turnSpeed = -engine->axis2X / 2000000.0;
		
		if ((engine->axis2X <= -9000) || (engine->axis2X >= 9000))
		{
			game->changeCameraRotation(turnSpeed * engine->getTimeDifference(TD_LOGIC));
		}
	}
	
	if (game->strafeLock)
	{
		if ((engine->axis2Y <= -game->joystickDeadZone) || (engine->axis2Y >= game->joystickDeadZone))
		{
			float turnSpeed = engine->axis2Y / 20000.0;
			
			if ((engine->axis2Y <= -9000) || (engine->axis2Y >= 9000))
			{
				game->aimDirection.y += turnSpeed;
			}
		}
	}
}

void movePlayer()
{
	float moveSpeed = 0.65;
	bool walkTurn = false;
	
	if (player->target != NULL)
	{
		player->target->referenced = true;
		
		if (Math::getDistance(player->position.x, player->position.y, player->target->position.x, player->target->position.y) > 250)
		{
			player->target = NULL;
		}
		else if (player->target->health <= 0)
		{
			player->target = NULL;
		}
		else if (!bsp->canSightTarget(player->position, player->target->position))
		{
			player->target = NULL;
		}
	}
	
	if (game->isControl(CTRL_LEFT))
	{
		player->velocity.x += (-0.25 * moveSpeed * -sin(game->cameraAdjustedRadian - 1.5));
		player->velocity.y += (-0.25 * moveSpeed * cos(game->cameraAdjustedRadian - 1.5));
		player->rotation.x = 270 + toDegrees(game->cameraAdjustedRadian);
		walkTurn = true;
	}
	
	if (game->isControl(CTRL_RIGHT))
	{
		player->velocity.x += (0.25 * moveSpeed * -sin(game->cameraAdjustedRadian - 1.5));
		player->velocity.y += (0.25 * moveSpeed * cos(game->cameraAdjustedRadian - 1.5));
		player->rotation.x = 90 + toDegrees(game->cameraAdjustedRadian);
		walkTurn = true;
	}
	
	if (game->isControl(CTRL_UP))
	{
		player->velocity.x += (0.25 * moveSpeed * -sin(game->cameraAdjustedRadian));
		player->velocity.y += (0.25 * moveSpeed * cos(game->cameraAdjustedRadian));
		player->rotation.x = 180 + toDegrees(game->cameraAdjustedRadian);
			
		if (game->isControl(CTRL_LEFT))
		{
			player->rotation.x += 45;
		}
		
		if (game->isControl(CTRL_RIGHT))
		{
			player->rotation.x -= 45;
		}
		
		walkTurn = true;
	}
	
	if (game->isControl(CTRL_DOWN))
	{
		player->velocity.x += (-0.25 * moveSpeed * -sin(game->cameraAdjustedRadian));
		player->velocity.y += (-0.25 * moveSpeed * cos(game->cameraAdjustedRadian));
		player->rotation.x = toDegrees(game->cameraAdjustedRadian);
		
		if (game->isControl(CTRL_LEFT))
		{
			player->rotation.x -= 45;
		}
		
		if (game->isControl(CTRL_RIGHT))
		{
			player->rotation.x += 45;
		}
		
		walkTurn = true;
	}
	
	if (game->isControl(CTRL_ROTATE_LEFT))
	{
		game->changeCameraRotation(0.025 * engine->getTimeDifference(TD_LOGIC));
	}
	
	if (game->isControl(CTRL_ROTATE_RIGHT))
	{
		game->changeCameraRotation(-0.025 * engine->getTimeDifference(TD_LOGIC));
	}
	
	if (game->isControl(CTRL_JUMP))
	{
		if ((player->liquidLevel == 0) && (player->flags & EF_ONGROUND))
		{
			player->velocity.z = 1.25;
			engine->keyState[SDLK_SPACE] = 0;
			Math::removeBit(&player->flags, EF_ONGROUND);
		}
		else if (player->liquidLevel > 0)
		{
			player->velocity.z += (0.01 * engine->getTimeDifference(TD_LOGIC));
		}
	}
	else if (player->liquidLevel > 0)
	{
		if (player->velocity.z > -0.5)
		{
			player->velocity.z -= (0.0025 * engine->getTimeDifference(TD_LOGIC));
		}
	}
	
	if (game->isControl(CTRL_STATUS))
	{
		game->resetControl(CTRL_STATUS);
		showMissionStatusPanel();
	}
	
	if (engine->keyState[SDLK_f])
	{
		player->velocity.z = 0;
		player->position.z += (engine->getTimeDifference(TD_LOGIC));
	}
	
	if (engine->keyState[SDLK_v])
	{
		player->velocity.z = 0;
		player->position.z -= (engine->getTimeDifference(TD_LOGIC));
	}
}

/*
 * Keeps the player a certain distance away from the edges by tracing down
 * in a large box around them. If it's possible to hit the target then the player
 * isn't in a safe place
*/
void assignCheckPoint()
{
	if (!(player->flags & EF_ONGROUND))
	{
		return;
	}
	
	Vector position = player->position;
	
	position = player->position;
	position.x -= 30;
	position.y -= 30;
	position.z -= 4;
	
	if (bsp->canSightTarget(player->position, position))
	{
		return;
	}
	
	position = player->position;
	position.x += 30;
	position.y -= 30;
	position.z -= 4;
	
	if (bsp->canSightTarget(player->position, position))
	{
		return;
	}
	
	position = player->position;
	position.x += 30;
	position.y += 30;
	position.z -= 4;
	
	if (bsp->canSightTarget(player->position, position))
	{
		return;
	}
	
	position = player->position;
	position.x -= 30;
	position.y += 30;
	position.z -= 4;
	
	if (bsp->canSightTarget(player->position, position))
	{
		return;
	}
	
	game->checkpoint = player->position;
}

void doPlayer()
{
	if (game->missionStatus == MISSION_INPROGRESS)
	{
		if (player->contentType & CONTENTS_SLIME)
		{
			player->applyDamage(4);
			player->helpless = RAND_MAX;
			game->setMissionStatus((player->health > 0) ? MISSION_PLAYER_OUT : MISSION_FAILED);
			game->alterStat("TIMES_FALLEN_IN_SLIME", 1);
		}
		else if (player->contentType & CONTENTS_LAVA)
		{
			player->applyDamage(8);
			player->helpless = RAND_MAX;
			game->setMissionStatus((player->health > 0) ? MISSION_PLAYER_OUT : MISSION_FAILED);
			game->alterStat("TIMES_FALLEN_IN_LAVA", 1);
		}
	}
	
	//player->health = player->maxHealth;
	
	if (!(player->flags & (EF_DYING|EF_DEAD)))
	{
		if (player->oxygen == 0)
		{
			if ((((int)graphics->getAnimTimer()) % 45) <= 0)
			{
				audio->playSound(SND_DROWNING, CH_PLAYER);
			}
		}
	}
	
	if ((player->health <= 0) || (player->action != NULL) || (player->helpless))
	{
		return;
	}
	
	assignCheckPoint();
	
	player->velocity.x *= 0.9;
	player->velocity.y *= 0.9;
	
	if ((player->velocity.x > -EPSILON) && (player->velocity.x < EPSILON))
	{
		player->velocity.x = 0;
	}
	
	if ((player->velocity.y > -EPSILON) && (player->velocity.y < EPSILON))
	{
		player->velocity.y = 0;
	}
	
	movePlayer();
	movePlayerAnalog();
	movePlayerMouse();
	
	if (game->strafeLock)
	{
		player->rotation.x = 180 + toDegrees(game->cameraAdjustedRadian);
		
		calculatePlayerAimDirection();
	
		drawLaserSight(player);
	}
	
	doPlayerWeapons();
	
	playerIssueOrders();
	
	Math::limit(&player->velocity.x, -1.2, 1.2);
	Math::limit(&player->velocity.y, -1.2, 1.2);
}
